<?php
if(!defined('PLX_ROOT')) { exit; }

require 'vendor/autoload.php';
use OTPHP\TOTP;

/**
 * A Two Factor Authentication, also known as 2FA plugin for PluXml, according to RFC 4226 (HOTP Algorithm) and RFC 6238 (TOTP Algorithm).
 * Based on https://github.com/Spomky-Labs/otphp library.
 *
 * @author J.P. Pourrez @bazooka07
 * @version 1.0.1
 * @link https://github.com/Spomky-Labs/otphp/blob/10.0.x/doc/index.md
 * */
class kzOtPHP extends plxPlugin {

	const HOOKS = array(
		# Authentification
		'PLX_AUTHPAGE'	=> array(
			'AdminAuthPrepend',
			'AdminAuth',
			'AdminAuthTop',
			'AdminAuthEndHead',
		),
		# User interface
		'PLX_ADMIN'		=> array(
			'AdminProfilTop',
			'AdminProfil',
			'AdminProfilFoot',
			'plxAdminEditProfil', # Manage the user
		),
	);
	const CODE_BEGIN = '<?php /* ' . __CLASS__ . ' plugin */' . PHP_EOL;
	const CODE_END = PHP_EOL . '?>';
	const FIELDNAME = 'otphp_enable';
	const AUTH_FIELDNAME = 'kz-otp';
	const DIGESTS = array('md5', 'sha1', 'sha256', 'sha512');
	const DEFAULT_DIGEST = 'sha1';
	const DEFAULT_PERIOD = 30;
	const WINDOW = 2;
	const LOGOS_ROOT = PLX_PLUGINS . __CLASS__ . '/logos/';

	const IMG_SIZE = 200; # for QR-code image

	const QR_CODE_PROVIDERS = array(
		'QR Server'		=> 'https://api.qrserver.com/v1/create-qr-code/?bgcolor=fff&data=[DATA]&qzone=2&margin=0&size=' . self::IMG_SIZE . 'x' . self::IMG_SIZE . '&ecc=M',
		'QRickit'		=> 'https://qrickit.com/api/qr.php?d=[DATA]&qrsize=' . self::IMG_SIZE,
		'QuickChart'	=> 'https://quickchart.io/qr?text=[DATA]&size=' . self::IMG_SIZE, # https://quickchart.io/documentation/qr-codes/
	);

	# Applications sur mobile pour générer le code Pin d'authentification
	const APPS = array(
		'F-Droid'	=> array(
			'storeUrl'	=> 'https://f-droid.org/packages/',
			'logo'		=> 'fdroid.svg',
			'apps'		=> array(
				'FreeOTP'	=> 'org.fedorahosted.freeotp',
				'Aegis'		=> 'com.beemdevelopment.aegis',
				'Secur'	=> 'com.u2fa.secur',
				'OneTimePass'	=> 'com.github.onetimepass',
			),
			'searchUrl'	=> 'https://search.f-droid.org/?q=2fa',
		),
		'Google'	=> array(
			'storeUrl'	=> 'https://play.google.com/store/apps/details?id=',
			'logo'		=> 'playstore.svg',
			'apps'		=> array(
				'LastPass'	=> 'com.lastpass.authenticator',
				'Vip_Access'	=> 'com.verisign.mvip.main',
				'Authy'	=> 'com.authy.authy',
				'Sophos'	=> 'com.sophos.sophtoken',
			),
			'searchUrl'	=> 'https://play.google.com/store/search?q=2fa&c=apps',
		),
	);

	# Use openssl_get_cipher_methods() PHP method to check if ENC_METHOS is available
	const ENCRYPT_METHOD = 'AES-128-CTR';

	public $success = null;
	public $allSecrets = null;
	private $label = null;
	private $__provider = null;

	public function __construct($default_lang) {

		parent::__construct($default_lang);

		foreach(self::HOOKS as $context=>$hooks) {
			if(defined($context)) {
				if($context == 'PLX_ADMIN') {
					# droits pour accéder à la page config.php du plugin
					parent::setConfigProfil(PROFIL_ADMIN);
				}

				foreach($hooks as $hook)
					parent::addHook($hook, $hook);
			}
		}
	}

	/**
	 * Clean up config file from old version.
	 * */
	function onActivate() {
		if(!empty(parent::getParam('context'))) {
			$this->delParam('context');
			$this->saveParams();
		}
	}

	private function __encrypt_key() {
		return openssl_digest(
			gethostname() . "|" . ip2long($_SERVER['SERVER_ADDR']),
			'SHA256',
			true
		);
	}

	/**
	 * the allsecrets param is a serialized array within the secret key for evey user,
	 * indexed by each user ID (3 digits as a string).
	 * */
	private function __getAllSecrets() {
		# http://www.the-art-of-web.com/php/two-way-encryption/
		$param = $this->getParam('allsecrets');
		if(!empty($param)) {
			# $this->allSecrets = unserialize($param);
			$enc_iv = hex2bin($this->getParam('iv'));
			$this->allSecrets = unserialize(openssl_decrypt(
				$param,
				self::ENCRYPT_METHOD,
				self::__encrypt_key(),
				0,
				$enc_iv
			));
			unset($enc_iv);
		} else {
			$this->allSecrets = array();
		}
	}

	private function __saveAllSecrets() {
		if(!empty($this->allSecrets)) {
			# $this->setParam('allsecrets', serialize($this->allSecrets), 'cdata');
			$enc_iv = openssl_random_pseudo_bytes(openssl_cipher_iv_length(self::ENCRYPT_METHOD));
			$this->setParam('iv', bin2hex($enc_iv), 'cdata');
			$this->setParam(
				'allsecrets',
				openssl_encrypt(
					serialize($this->allSecrets),
					self::ENCRYPT_METHOD,
					self::__encrypt_key(),
					0,
					$enc_iv
				),
				'cdata'
			);
			unset($enc_iv);
		} else {
			$this->delParam('iv');
			$this->delParam('allsecrets');
			$this->delParam('context');
		}
		$this->saveParams();
	}

	private function __createTOTP($secret=null) {
		$digest = $this->getParam('digest');
		if(empty($digest) or !in_array($digest, self::DIGESTS)) {
			$digest = self::DEFAULT_DIGEST;
		}
		$digits = intval($this->getParam('digits'));
		if($digits < 3 or $digits > 10) { $digits = 6; }
		// $this->__getAllSecrets();
		if(false) {
			# for oldest versions of TOTP
			return new TOTP(
			    $this->label,
			    $secret, # $this->allSecrets[$userId]
			    self::DEFAULT_PERIOD, # period
			    $digest, # sha1, sha256, md5, ...
			    $digits # number of digits
			);
		} else {
			# from TOTP version 10.0.0
			$totp = TOTP::create(
				$secret, # $this->allSecrets[$userId]
			    self::DEFAULT_PERIOD, # period
			    $digest, # sha1, sha256, md5, ...
			    $digits # number of digits
			);
			$totp->setLabel($this->label);
			return $totp;
		}
	}

	/**
	 * This is the core of double HOTP and TOTP.
	 * */
	public function check(&$plxAdmin, $counter) {
		# No hook in PluXml after authentification is up !!!
		# So, we do the same job but in a better way.
		if(empty($_POST['login']) or empty($_POST['password'])) {
			return;
		}

		$params = filter_input_array(INPUT_POST, array(
			'login'			=> FILTER_SANITIZE_SPECIAL_CHARS,
			'password'		=> FILTER_SANITIZE_SPECIAL_CHARS,
			self::AUTH_FIELDNAME	=> array(
				'filter' => FILTER_VALIDATE_REGEXP,
				'options'=> array('regexp'=> '@^\d{3,10}$@')
			)
		));

		if(
			!empty($params['login']) and
			!empty($params['password'])
		) {
			$this->__getAllSecrets();
			if(!empty($this->allSecrets)) {
				$pass_md5 = md5($params['password']);
				foreach($plxAdmin->aUsers as $userId => $user) {
					if(
						!empty($user['active']) and !$user['delete'] and
						$_POST['login'] === $user['login'] and
						sha1($user['salt'].$pass_md5) === $user['password']
					) {
						# This user is found.
						if(array_key_exists($userId, $this->allSecrets)) {
							# This user is using the double authentification.
							$this->success = false;
							if(!empty($params[self::AUTH_FIELDNAME])) {
								$this->label = $plxAdmin->aUsers[$userId]['email'];
								$totp = $this->__createTOTP($this->allSecrets[$userId]);
								$this->success = $totp->verify($params[self::AUTH_FIELDNAME], null, self::WINDOW);
							}
							if(!$this->success) {
								# Authentification fails ! So desactivate the user.
								$plxAdmin->aUsers[$userId]['active'] = false;
							}
						}
						break;
					}
				}
			}
		}
	}

	/**
	 * Callback for AdminProfilTop() hook.
	 * */
	public function setSecret(&$plxAdmin) {
		$this->__getAllSecrets();
		$userId = $_SESSION['user'];
		if(!empty($this->allSecrets) and array_key_exists($userId, $this->allSecrets)) {
			$login = $plxAdmin->aUsers[$userId]['login'];
			$this->secret = $this->allSecrets[$userId];
			$this->label = "{$login}@{$_SERVER['HTTP_HOST']}";
			$this->issuer = $plxAdmin->aConf['title'];
		}
	}

	/**
	 * Callback for plxAdminEditProfil() hook.
	 *
	 * Creates a secret for the current user.
	 * Also stores title of the page and login for using with internal QR-Code builder.
	 * */
	public function createSecret(&$plxAdmin) {
		$this->__getAllSecrets();
		$userId = $_SESSION['user'];
		if(!array_key_exists($userId, $this->allSecrets)) {
			$this->label = $plxAdmin->aUsers[$userId]['email'];
			$totp = $this->__createTOTP();
			$this->allSecrets[$userId] = $totp->getSecret();
			$this->__saveAllSecrets();
		}
	}

	/**
	 * 2nd Callback for plxAdminEditProfil() hook.
	 * */
	public function delSecret() {
		$this->__getAllSecrets();
		$userId = $_SESSION['user'];
		if(!empty($this->allSecrets) and array_key_exists($userId, $this->allSecrets)) {
			unset($this->allSecrets[$userId]);
			$this->__saveAllSecrets();
		}
	}

	/**
	 * Builds QR-code and returns as a string.
	 * */
	private function __buildQRcode_img($provisioningUri) {
		$renderer = new \BaconQrCode\Renderer\ImageRenderer(
		    new \BaconQrCode\Renderer\RendererStyle\RendererStyle(self::IMG_SIZE),
		    new \BaconQrCode\Renderer\Image\SvgImageBackEnd()
		);
		$writer = new \BaconQrCode\Writer($renderer);
		$xml = $writer->writeString($provisioningUri);
		# On supprime l'entête XML
		return preg_replace('#^<\?xml [^>]*>\s*#', '', $xml);
	}

/* ====================== Hooks ============================ */

	/**
	 * Use $this->check() for checking the code PIN.
	 * Desactivate the user if double-authentification fails when login and password match.
	 * */
	public function AdminAuthPrepend() {
		echo self::CODE_BEGIN;
?>
if(!isset($maxlogin)) {
	$maxlogin = array(
		'counter'	=> MAX_LOGIN_COUNT,
		'timer'		=> MAX_LOGIN_TIME,
	);
}
if(
	empty($_SESSION['maxtry']['counter']) or
	empty($_SESSION['maxtry']['timer']) or
	intval($_SESSION['maxtry']['counter']) < $maxlogin['counter'] or
	time() <= $_SESSION['maxtry']['timer'] + $maxlogin['timer']
) {
	$plxAdmin->plxPlugins->aPlugins['<?= __CLASS__ ?>']->check($plxAdmin, $maxlogin['counter']);
}
<?php
		echo self::CODE_END;
	}

	/**
	 * Insère une entrée dans le formulaire pour saisir le code Pin.
	 * */
	public function AdminAuth() {
		$name = self::AUTH_FIELDNAME;
		$caption = parent::getLang('OTP_PIN');
?>
		<div class="grid">
			<div class="col sml-12">
				<input id="id_<?= $name ?>" type="text" name="<?= $name ?>" value="" class="full-width text-center" maxlength="10" pattern="^\d{3,10}$" placeholder="<?= $caption ?>" inputmode="numeric" autocomplete="off" />
			</div>
		</div>
<?php
	}

	/**
	 * Update $msg if no success.
	 * */
	public function AdminAuthTop() {
		if(isset($this->success) and $this->success === false) {
			echo self::CODE_BEGIN;
?>
if(empty($msg) or $msg == L_ERR_WRONG_PASSWORD) {
	$msg = '<?= $this->getLang('OTP_FAILURE') ?>';
	$error = 'error';
}
<?php
			echo self::CODE_END;
		}
	}

	public function AdminAuthEndHead() {
?>
	<style type="text/css">
		body { background-image: url('<?= PLX_PLUGINS. __CLASS__ . '/darkness-100.png' ?>'); }
		section > div.auth { background-color: #f0f0f0; }
	</style>
<?php
	}
	/* -------------------- Setup for the current user ---------------------- */

	/**
	 * Get the secret, Active the cache for output.
	 *
	 * Calls $this->setScecret(...)
	 * */
	public function AdminProfilTop() {
		echo self::CODE_BEGIN;
?>
$plxAdmin->plxPlugins->aPlugins['<?= __CLASS__ ?>']->setSecret($plxAdmin);
ob_start(); # On démarre la bufférisation pour ajouter un champ dans le formulaire
<?php
		echo self::CODE_END;
	}

	/**
	 * Add the select tag for the double authentification for the current user.
	 * */
	public function AdminProfil() {
		ob_start();
?>
		<div class="grid">
			<div class="col sml-12">
				<label for="id_otphp_enable"><?= $this->getLang('OTP_ENABLE') ?>&nbsp;:</label>
				<?php plxUtils::printSelect(self::FIELDNAME, array('1'=>L_YES,'0'=>L_NO), empty($this->secret) ? '0' : '1'); ?>
			</div>
		</div>
<?php
		$replace = ob_get_clean();
		echo self::CODE_BEGIN;
?>
$kzOutput = ob_get_clean(); # Fin de la bufférisation lancée par le hook AdminProfilTop
echo preg_replace('@(\s*</fieldset>.*)@s', '<?= $replace ?>$1', $kzOutput);
<?php
		echo self::CODE_END;
	}

	/**
	 * Checks if the internal QR-code builder can be using.
	 * */
	private function __use_internal_provider() {
		$this->__provider = $this->getParam('qrcode_provider');
		if(!empty($this->__provider) and array_key_exists($this->__provider, self::QR_CODE_PROVIDERS)) {
			# With Alpinelinux, php-iconv and ctype aren't installed by default.
			if(!function_exists('iconv')) {
				plxMsg::Error($this->getLang('MISSING_ICONV_LIBRARY'));
				return true;
			}
			if(!function_exists('ctype_digit')) {
				plxMsg::Error($this->getLang('MISSING_CTYPE_LIBRARY'));
				return true;
			}
			return false;
		}
		return true;
	}

	/**
	 * Display the secret with a QR-Code.
	 * */
	public function AdminProfilFoot() {
		if(!empty($this->secret)) {
			$totp = $this->__createTOTP($this->secret);
			$totp->setIssuer($this->issuer);
			$provisioningUri = $totp->getProvisioningUri();
			if(self::__use_internal_provider()) {
				# image vectorielle
				$imgData = self::__buildQRcode_img($provisioningUri);
				$src = 'data:image/svg+xml,' . rawurlencode($imgData);
			} else {
				$src = $totp->getQrCodeUri(self::QR_CODE_PROVIDERS[$this->__provider], '[DATA]');
			}
?>
	<figure class="<?= __CLASS__; ?> qrcode">
		<p><?= implode(' - ', array($this->getParam('digest'), $this->getParam('digits').' digits')); ?></p>
		<img id="qr-code" alt="QR-Code" width="<?= SELF::IMG_SIZE; ?>" height="<?= SELF::IMG_SIZE; ?>"
			src="<?= $src; ?>"
			title="<?= $provisioningUri; ?>"
		/>
		<figcaption><?php $this->lang('OTP_SCAN'); ?></figcaption>
	</figure>
	<input type="text" id="clipboard" class="<?= __CLASS__; ?> off-screen" />
	<script type="text/javascript">
		(function() {
			'use strict';
			const qrCode = document.getElementById('qr-code');
			const clp = document.getElementById('clipboard');
			if(qrCode != null && clp != null) {
				qrCode.addEventListener('click', function(event) {
					clp.value = event.target.title;
					clp.select();
					document.execCommand('copy');
					event.preventDefault();
					alert('<?php $this->lang("COPIED"); ?>');
				})
			}
		})();
	</script>
<?php
		}
	}

	/**
	 * Generate the secret if required,
	 * Otherwise delete the secret if desactivation.
	 * */
	public function plxAdminEditProfil() {
		echo self::CODE_BEGIN;
?>
if(array_key_exists('<?= self::FIELDNAME ?>', $content)) {
	if($content['<?= self::FIELDNAME ?>'] == '1') {
		$this->plxPlugins->aPlugins['<?= __CLASS__ ?>']->createSecret($this);
	} elseif($content['<?= self::FIELDNAME ?>'] == '0') {
		$this->plxPlugins->aPlugins['<?= __CLASS__ ?>']->delSecret();
	}
}
<?php
		echo self::CODE_END;
	}

}
